/*
 * Copyright (c) 2004,2005 Voltaire Inc.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: smpquery.c 3541 2005-09-26 13:47:49Z halr $
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>

#include <common.h>
#include <umad.h>
#include <mad.h>

#undef DEBUG
#define	DEBUG	if (verbose>1) IBWARN
#define IBERROR(fmt, args...)	iberror(__FUNCTION__, fmt, ## args)

static int dest_type = IB_DEST_LID;
static int verbose;

typedef char *(op_fn_t)(ib_portid_t *dest, char **argv, int argc);

typedef struct match_rec {
	char *name;
	op_fn_t *fn;
} match_rec_t;

static op_fn_t	node_desc, node_info, port_info, switch_info;

static match_rec_t match_tbl[] = {
	{ "nodeinfo", node_info },
	{ "nodedesc", node_desc },
	{ "portinfo", port_info },
	{ "switchinfo", switch_info },
	{0}
};

static void
iberror(const char *fn, char *msg, ...)
{
	char buf[512], *s;
	va_list va;
	int n;

	va_start(va, msg);
	n = vsprintf(buf, msg, va);
	va_end(va);
	buf[n] = 0;

	if ((s = strrchr(argv0, '/')))
		argv0 = s + 1;

	if (verbose)
		printf("%s: iberror: [pid %d] %s: %s\n", argv0, getpid(), fn, buf);
	else
		printf("%s: iberror: %s\n", argv0, buf);

	exit(-1);
}

/*******************************************/

static char *
node_desc(ib_portid_t *dest, char **argv, int argc)
{
	char nd[IB_SMP_DATA_SIZE];

	if (!smp_query(nd, dest, IB_ATTR_NODE_DESC, 0, 0))
		return "smp query failed";

	_mad_dump(mad_dump_nodedesc, "Node Description", nd, sizeof nd);
	return 0;
}

static char *
node_info(ib_portid_t *dest, char **argv, int argc)
{
	char buf[2048];
	char data[IB_SMP_DATA_SIZE];

	if (!smp_query(data, dest, IB_ATTR_NODE_INFO, 0, 0))
		return "smp query failed";

	mad_dump_nodeinfo(buf, sizeof buf, data, sizeof data);

	printf("# Node info: %s\n%s", portid2str(dest), buf);
	return 0;
}

static char *
port_info(ib_portid_t *dest, char **argv, int argc)
{
	char buf[2048];
	char data[IB_SMP_DATA_SIZE];
	int portnum = 0;

	if (argc > 0)
		portnum = strtol(argv[0], 0, 0);

	if (!smp_query(data, dest, IB_ATTR_PORT_INFO, portnum, 0))
		return "smp query failed";

	mad_dump_portinfo(buf, sizeof buf, data, sizeof data);

	printf("# Port info: %s port %d\n%s", portid2str(dest), portnum, buf);
	return 0;
}

static char *
switch_info(ib_portid_t *dest, char **argv, int argc)
{
	char buf[2048];
	char data[IB_SMP_DATA_SIZE];

	if (!smp_query(data, dest, IB_ATTR_SWITCH_INFO, 0, 0))
		return "smp query failed";

	mad_dump_switchinfo(buf, sizeof buf, data, sizeof data);

	printf("# Switch info: %s\n%s", portid2str(dest), buf);
	return 0;
}

op_fn_t *
match_op(char *name)
{
	match_rec_t *r;

	for (r = match_tbl; r->name; r++)
		if (!strcmp(r->name, name))
			return r->fn;

	return 0;
}

void
usage(void)
{
	char *basename;

	if (!(basename = strrchr(argv0, '/')))
		basename = argv0;
	else
		basename++;

	fprintf(stderr, "Usage: %s [-d(ebug) -e(rr_show) -v(erbose) -D(irect) -G(uid) -s smlid -V(ersion) -C ca_name -P hca_port "
			"-t timeout_ms] <op> <dest dr_path|lid|guid> [op params]\n",
			basename);
	fprintf(stderr, "\tsupported ops:\n");
	fprintf(stderr, "\t\tnodeinfo <addr>\n");
	fprintf(stderr, "\t\tnodedesc <addr>\n");
	fprintf(stderr, "\t\tportinfo <addr> [<portnum>]\n");
	fprintf(stderr, "\t\tswitchinfo <addr>\n");
	fprintf(stderr, "\n\texamples:\n");
	fprintf(stderr, "\t\t%s portinfo 3 1\t\t\t\t# switchinfo by lid, with port modifier\n", basename);
	fprintf(stderr, "\t\t%s -G switchinfo 0x2C9000100D051 1\t# switchinfo by guid\n", basename);
	fprintf(stderr, "\t\t%s -D nodeinfo 0\t\t\t# nodeinfo by direct route\n", basename);
	exit(-1);
}

int
main(int argc, char **argv)
{
	int mgmt_classes[3] = {IB_SMI_CLASS, IB_SMI_DIRECT_CLASS, IB_SA_CLASS};
	ib_portid_t portid = {0};
	ib_portid_t *sm_id = 0, sm_portid = {0};
	extern int ibdebug;
	int timeout = 0, udebug = 0;
	char *ca = 0;
	int ca_port = 0;
	char *err;
	op_fn_t *fn;

	ARGBEGIN {
	case 'd':
		ibdebug++;
		madrpc_show_errors(1);
		umad_debug(udebug);
		udebug++;
		break;
	case 'e':
		madrpc_show_errors(1);
		break;
	case 'D':
		dest_type = IB_DEST_DRPATH;
		break;
	case 'G':
		dest_type = IB_DEST_GUID;
		break;
	case 'C':
		ca = ARGF();
		break;
	case 'P':
		ca_port = strtoul(ARGF(), 0, 0);
		break;
	case 's':
		if (ib_resolve_portid_str(&sm_portid, ARGF(), IB_DEST_LID, 0) < 0)
			IBERROR("can't resolve SM destination port %s", ARGF());
		sm_id = &sm_portid;
		break;
	case 't':
		timeout = strtoul(ARGF(), 0, 0);
		madrpc_set_timeout(timeout);
		break;
	case 'v':
		verbose++;
		break;
	case 'V':
		fprintf(stderr, "%s %s\n", argv0, get_build_version() );
		exit(-1);
	case 'h':
	default:
		usage();
	} ARGEND;

	if (argc < 2)
		usage();

	if (!(fn = match_op(argv[0])))
		IBERROR("operation '%s' not supported", argv[0]);

	madrpc_init(ca, ca_port, mgmt_classes, 3);

	if (ib_resolve_portid_str(&portid, argv[1], dest_type, sm_id) < 0)
		IBERROR("can't resolve destination port %s", argv[1]);

	if ((err = fn(&portid, argv+2, argc-2)))
		IBERROR("operation %s failed: %s", argv[0], err);

	exit(0);
}
